#ifndef __RawServerSocket_h__
#define __RawServerSocket_h__

#include "Error.h"
#include "ISelector.h"
#include "T/SocketFactory.h"
#include "T/sys/socket.h"
#include "T/unistd.h"
#include <arpa/inet.h>
#include <string.h>
#include <sys/types.h>

#include <cstdio>

class RawServerSocket : public ISelector::ISelectable
{
public:
	enum {
		MAX_NUMBER_OF_PENDING_CONNECTIONS = 10,
	};

	RawServerSocket( ISelector &selector ) : _descriptor( 0 ), _selector( selector )
	{
		createSocket();
        callBind();

		registerToSelector();

                testCheckSum();
	}

	~RawServerSocket()
	{
		unregisterFromSelector();
		closeSocketIfNeeded();
	}

	void canRead()
	{
        const int MAX_DATA_SIZE = 10000;
        const int NO_FLAGS = 0;

        char buffer[MAX_DATA_SIZE];
        size_t actualReceivedBytes = sizeof(buffer);

        struct sockaddr sourceAddress;
        memset( &sourceAddress, 0, sizeof(sourceAddress) );

        socklen_t addressLength = sizeof(sourceAddress);

        actualReceivedBytes = recvfrom( _descriptor, buffer, sizeof(buffer), NO_FLAGS, &sourceAddress, &addressLength );

        struct sockaddr_in *asIp = (struct sockaddr_in *)&sourceAddress;

        printf( "Recevied from ip: %s, port: %u\n", inet_ntoa( asIp->sin_addr ), ntohs( asIp->sin_port ) );
        fflush( 0 );

        //         address.sin_family = AF_INET;
        // address.sin_addr.s_addr = ip;
        // address.sin_port = htons( port );


// ssize_t recvfrom(int sockfd, void *buf, size_t len, int flags,
//                         struct sockaddr *src_addr, socklen_t *addrlen);

//        actualReceivedBytes = T::recv( _descriptor, buffer, sizeof(buffer), NO_FLAGS );

        if ( actualReceivedBytes == 0 )
        {
            throw Error( Error::SOCKET_WAS_CLOSED );
        }
        else if ( (int)actualReceivedBytes == -1 )
        {
            throw Error( Error::RECV_FAILED );
        }

        ConstSimpleData( buffer, actualReceivedBytes ).hexDump();

        char synAck[40] = { 0x02, 0x2b, // Source Port
                            buffer[20], buffer[21], // Destination Port
                            0xfc, 0xd2, 0xed, 0x9b, // Sequence number
                            buffer[24], buffer[25], buffer[26], buffer[27] + 1, // Ack number
                            0xa0, // header length
                            0x12, // flags: syn ack
                            0x80, 0x00, //window ssize
                            0x00, 0x00, // checksum
                            0x00, 0x00,
                            // Options
                            0x02, 0x04, 0x40, 0x0c, 0x04, 0x02, 0x08, 0x0a,
                            buffer[48], buffer[49], buffer[50], buffer[51],
                            buffer[48], buffer[49], buffer[50], buffer[51],
                            // 0x00, 0x24, 0xed, 0x46,
                            // 0x00, 0x24, 0xed, 0x46,
                            0x01, 0x03, 0x03, 0x04 };

        unsigned char checkSum[2];


        unsigned char pseudoHeader[12] =
            {
                0x7f, 0x00, 0x00, 0x01,
                0x7f, 0x00, 0x00, 0x01,
                0x00, 0x06, 0x00, 0x28
            };

        unsigned forChecksum[52];
        memcpy( forChecksum, pseudoHeader, sizeof(pseudoHeader) );
        memcpy( forChecksum + sizeof(pseudoHeader), synAck, sizeof(synAck) );

        *(unsigned short *)checkSum = comp_chksum( (unsigned short *)forChecksum, sizeof(forChecksum) );

        synAck[16] = checkSum[0];
        synAck[17] = checkSum[1];

        struct sockaddr destinationAddress;
        memset( &destinationAddress, 0, sizeof(destinationAddress) );

        memcpy( &destinationAddress, &sourceAddress, sizeof(struct sockaddr) );

        if ( sendto( _descriptor, synAck, sizeof(synAck), NO_FLAGS, &destinationAddress, sizeof(destinationAddress) ) != sizeof(synAck) )
        {
            throw Error( Error::SEND_FAILED );
        }

                sleep( 5 );
               // ssize_t sendto(int sockfd, const void *buf, size_t len, int flags,
               //        const struct sockaddr *dest_addr, socklen_t addrlen);


    }

    void send( const ConstSimpleData &data )
    {
        const int NO_FLAGS = 0;

        if ( T::send( _descriptor, data.data(), data.size(), NO_FLAGS ) != (int)data.size() )
        {
            throw Error( Error::SEND_FAILED );
        }
    }

	int descriptor()
	{
		return _descriptor;
	}

private:
	int _descriptor;
	ISelector &_selector;

	void registerToSelector()
	{
		_selector.registerToSelector( *this );
	}

	void unregisterFromSelector()
	{
		_selector.unregisterFromSelector( *this );
	}

	void createSocket()
	{
        if ( ( _descriptor = T::socket( AF_INET, SOCK_RAW, IPPROTO_TCP ) ) == -1 )
		{
			throw Error( Error::SOCKET_CREATION_FAILED );
		}
	}

	void callBind()
	{
		struct sockaddr_in socketAddress;
		memset( &socketAddress, 0, sizeof(struct sockaddr_in) );

		socketAddress.sin_family = AF_INET;
        //socketAddress.sin_addr.s_addr = INADDR_ANY;

        if ( inet_aton( "127.0.0.1", &socketAddress.sin_addr ) == 0 )
        {
            printf( "Invalid address\n" );
            exit( 1 );
        }

        if ( T::bind( _descriptor, (struct sockaddr *)&socketAddress, sizeof(struct sockaddr_in) ) == -1 )
		{
			closeSocket();
			throw Error( Error::BIND_FAILED );
		}
	}

	void closeSocketIfNeeded()
	{
		if ( _descriptor > 0 )
		{
			closeSocket();
		}
	}

	void closeSocket()
	{
		T::close( _descriptor );
		_descriptor = 0;
	}

    unsigned short comp_chksum( unsigned short *addr, int len )
    {
        long sum = 0;

        while (len > 1) {
            sum += *(addr++);
            len -= 2;
        }

        if (len > 0)
            sum += *addr;

        while (sum >> 16)
            sum = ((sum & 0xffff) + (sum >> 16));

        sum = ~sum;

        return ((u_short) sum);
    }

    void testCheckSum()
    {
        char buffer[] =
            {
                0x02, 0x2b, 0x85, 0xd0, 0xfc, 0xd2, 0xed, 0x9b,
                0x23, 0x29, 0x96, 0x3d, 0xa0, 0x12, 0x80, 0x00,
                0x00, 0x00, //0x97, 0x90,
                0x00, 0x00, 0x02, 0x04, 0x40, 0x0c,
                0x04, 0x02, 0x08, 0x0a, 0x00, 0x47, 0xe4, 0xec,
                0x00, 0x47, 0xe4, 0xec, 0x01, 0x03, 0x03, 0x04,



                // 0x7f, 0x00, 0x00, 0x01,
                // 0x7f, 0x00, 0x00, 0x01,
                // 0x00, 0x06, 0x00, 0x28,

                // 0x02, 0x2b, 0x82, 0xf3, 0xcf, 0xfb, 0xaf, 0x47,
                // 0xd0, 0xba, 0xc5, 0x78, 0xa0, 0x12, 0x80, 0x00,
                // 0x00, 0x00, //0xfe, 0x30,
                // 0x00, 0x00, 0x02, 0x04, 0x40, 0x0c,
                // 0x04, 0x02, 0x08, 0x0a, 0x00, 0x39, 0x8d, 0x45,
                // 0x00, 0x39, 0x8d, 0x45, 0x01, 0x03, 0x03, 0x04
            };

        unsigned char checkSum[2];

        *(unsigned short *)checkSum = comp_chksum( (unsigned short *)buffer, sizeof(buffer) );

        printf( "Check sum: 0x%02x 0x%02x\n", checkSum[0], checkSum[1] );
        fflush( 0 );

    }

};

#endif // __RawServerSocket_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
