#ifndef __AbstractStateSplitter_h__
#define __AbstractStateSplitter_h__

#include "AbstractionData.h"
#include "MultipleThreadIntersector.h"
#include "String.h"
#include "ThreadState.h"
#include "Vector.h"

class AbstractStateSplitter
{
public:
    struct SplittingInformation
    {
        SplittingInformation( unsigned moduleIndex, const String &stateToSplit, const Vector<String> &secedingConcreteStates ) :
            _moduleIndex( moduleIndex ), _stateToSplit( stateToSplit ), _secedingConcreteStates( secedingConcreteStates )
        {
        }

        unsigned _moduleIndex;
        String _stateToSplit;
        Vector<String> _secedingConcreteStates;
    };

    static void splitState( AbstractionData &abstractionData, SplittingInformation &splittingInformation )
    {
        Vector<SplittingInformation> asVector;
        asVector.append( splittingInformation );
        splitState( abstractionData, asVector );
    }

    static void splitState( AbstractionData &abstractionData, Vector<SplittingInformation> &splittingInformation )
    {
        AbstractStateSplitter( abstractionData, splittingInformation ).run();
    }

private:
    AbstractionData &_abstractionData;
    unsigned _moduleIndex;
    String _stateToSplit;
    Vector<String> _secedingConcreteStates;
    Vector<String> _remainingConcreteStates;
    String _nameOfNewSecedingState;
    String _nameOfNewRemainingState;
    Vector<SplittingInformation> _splittingInformation;

    AbstractStateSplitter( AbstractionData &abstractionData, Vector<SplittingInformation> splittingInformation ) :
        _abstractionData( abstractionData ), _splittingInformation( splittingInformation )
    {
    }

    void run()
    {
        for ( unsigned i = 0; i < _splittingInformation.size(); ++i )
        {
            initSplittingRequest( i );
            print();
            calculateRemainingStates();
            nameNewStates();
            createNewStates();

            deleteOriginalState();
            updateInitialState();
        }

        updateComposedGraph();
    }

    void initSplittingRequest( unsigned requestIndex )
    {
        _moduleIndex = _splittingInformation[requestIndex]._moduleIndex;
        _stateToSplit = _splittingInformation[requestIndex]._stateToSplit;
        _secedingConcreteStates = _splittingInformation[requestIndex]._secedingConcreteStates;

        _remainingConcreteStates.clear();
        _nameOfNewRemainingState = _nameOfNewSecedingState = "";
    }

    void print()
    {
        printf( "Working on module #%u\n", _moduleIndex );
        printf( "State to split: %s\n", _stateToSplit.ascii() );
        printf( "Concrete states to form a new state\n" );
        for ( unsigned i = 0; i < _secedingConcreteStates.size(); ++i )
        {
            printf( "%s\n", _secedingConcreteStates[i].ascii() );
        }

        printf( "\n" );
        fflush( 0 );
    }

    void calculateRemainingStates()
    {
        Vector<String> allConcreteStates = _abstractionData._abstractModules[_moduleIndex]._abstractionFunction->abstractToConcrete( _stateToSplit );

        for ( unsigned i = 0; i < allConcreteStates.size(); ++i )
        {
            if ( !_secedingConcreteStates.exists( allConcreteStates[i] ) )
            {
                _remainingConcreteStates.append( allConcreteStates[i] );
            }
        }
    }

    void nameNewStates()
    {
        _nameOfNewSecedingState = _stateToSplit + "_1";
        _nameOfNewRemainingState = _stateToSplit + "_2";
    }

    void createNewStates()
    {
        updateAbstractionFunction();
        updateAbstractModule();
    }

    void updateAbstractionFunction()
    {
        for ( unsigned i = 0; i < _secedingConcreteStates.size(); ++i )
        {
            _abstractionData._abstractModules[_moduleIndex]._abstractionFunction->add( _secedingConcreteStates[i], _nameOfNewSecedingState );
        }

        for ( unsigned i = 0; i < _remainingConcreteStates.size(); ++i )
        {
            _abstractionData._abstractModules[_moduleIndex]._abstractionFunction->add( _remainingConcreteStates[i], _nameOfNewRemainingState );
        }
    }

    void updateAbstractModule()
    {
        ThreadState state;
        ThreadState complementState;

        state._label = _nameOfNewSecedingState;
        complementState._label = _nameOfNewRemainingState;

        handleEventSets( state, _secedingConcreteStates );
        handleEventSets( complementState, _remainingConcreteStates );

        handleTransitions( state, _secedingConcreteStates );
        handleTransitions( complementState, _remainingConcreteStates );

        _abstractionData._abstractModules[_moduleIndex]._abstract->_graph[_nameOfNewSecedingState] = state;
        _abstractionData._abstractModules[_moduleIndex]._abstract->_graph[_nameOfNewRemainingState] = complementState;
    }

    void handleEventSets( ThreadState &abstractState, Vector<String> &concreteStates )
    {
        if ( concreteStates.empty() )
        {
            TS_TRACE( "Error!!!" );
        }

        abstractState._requestedEvents = _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteStates[0]]._requestedEvents;
        abstractState._blockedEvents = _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteStates[0]]._blockedEvents;
        abstractState._atomicPropositions = _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteStates[0]]._atomicPropositions;

        for ( unsigned i = 1; i < concreteStates.size(); ++i )
        {
            for ( unsigned j = 0; j < _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteStates[i]]._requestedEvents.size(); ++j )
            {
                if ( !abstractState._requestedEvents.exists( _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteStates[i]]._requestedEvents[j] ) )
                {
                    abstractState._requestedEvents.append( _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteStates[i]]._requestedEvents[j] );
                }
            }

            Vector<Event> copyOfBlocked = abstractState._blockedEvents;

            for ( unsigned j = 0; j < copyOfBlocked.size(); ++j )
            {
                if ( !_abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteStates[i]]._blockedEvents.exists( copyOfBlocked[j] ) )
                {
                    abstractState._blockedEvents.eraseByValue( copyOfBlocked[j] );
                }
            }

            abstractState._atomicPropositions.add( _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteStates[i]]._atomicPropositions );
        }
    }

    void handleTransitions( ThreadState &abstractState, Vector<String> &concreteStates )
    {
        for ( unsigned i = 0; i < concreteStates.size(); ++i )
        {
            handleTransitionsOfState( abstractState, concreteStates[i] );
        }
    }

    void handleTransitionsOfState( ThreadState &abstractState, String &concreteState )
    {
        for ( unsigned i = 0; i < _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteState]._successors.size(); ++i )
        {
            handleSingleSuccessorTransition( abstractState, _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteState]._successors[i] );
        }

        for ( unsigned i = 0; i < _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteState]._predecessors.size(); ++i )
        {
            handleSinglePredecessorTransition( abstractState, _abstractionData._abstractModules[_moduleIndex]._concrete->_graph[concreteState]._predecessors[i] );
        }
    }

    void handleSingleSuccessorTransition( ThreadState &abstractState, Transition &transition )
    {
        String targetAbstractState = _abstractionData._abstractModules[_moduleIndex]._abstractionFunction->concreteToAbstract( transition._state );

        Transition abstractSuccessorTransition( transition._event, targetAbstractState );
        Transition abstractPredecessorTransition( transition._event, abstractState._label );

        if ( !abstractState._successors.exists( abstractSuccessorTransition ) )
        {
            abstractState._successors.append( abstractSuccessorTransition );
            _abstractionData._abstractModules[_moduleIndex]._abstract->_graph[targetAbstractState]._predecessors.append( abstractPredecessorTransition );
        }
    }

    void handleSinglePredecessorTransition( ThreadState &abstractState, Transition &transition )
    {
        String sourceAbstractState = _abstractionData._abstractModules[_moduleIndex]._abstractionFunction->concreteToAbstract( transition._state );

        Transition abstractPredecessorTransition( transition._event, sourceAbstractState );
        Transition abstractSuccessorTransition( transition._event, abstractState._label );

        if ( !abstractState._predecessors.exists( abstractPredecessorTransition ) )
        {
            abstractState._predecessors.append( abstractPredecessorTransition );
            _abstractionData._abstractModules[_moduleIndex]._abstract->_graph[sourceAbstractState]._successors.append( abstractSuccessorTransition );
        }
    }

    void deleteOriginalState()
    {
        deleteFromAbstractionFunction();
        deleteFromModule();
    }

    void deleteFromAbstractionFunction()
    {
        _abstractionData._abstractModules[_moduleIndex]._abstractionFunction->deleteAbstractState( _stateToSplit );
    }

    void deleteFromModule()
    {
        _abstractionData._abstractModules[_moduleIndex]._abstract->removeState( _stateToSplit );
    }

    void updateComposedGraph()
    {
        _abstractionData._composedGraph->clear();

        MultipleThreadIntersector intersector;

        for ( unsigned i = 0; i < _abstractionData._abstractModules.size(); ++i )
        {
            intersector.addThread( _abstractionData._abstractModules[i]._abstract );
        }

        intersector.intersect( *_abstractionData._composedGraph, ' ' );

        _abstractionData._composedGraph->applyAxioms();
    }

    void updateInitialState()
    {
        _abstractionData._abstractModules[_moduleIndex]._abstract->_initialState =
            _abstractionData._abstractModules[_moduleIndex]._abstractionFunction->concreteToAbstract( _abstractionData._abstractModules[_moduleIndex]._concrete->_initialState );
    }
};

#endif // __AbstractStateSplitter_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
