#ifndef __DirectAbstraction_h__
#define __DirectAbstraction_h__

#include "AbstractionData.h"
#include "NFA.h"

class DirectAbstraction
{
public:
    typedef Vector<String> StateSet;

    struct AbstractStateData
    {
        AbstractStateData( Vector<String> &concreteStates, String abstractState ) : _concreteStates( concreteStates ), _abstractState( abstractState )
        {
        }

        Vector<String> _concreteStates;
        String _abstractState;

        void dump()
        {
            printf( "New abstract state: %s\n", _abstractState.ascii() );
            printf( "Concrete states: \n" );

            for ( unsigned i = 0; i < _concreteStates.size(); ++i )
            {
                printf( "\t%s\n", _concreteStates[i].ascii() );
            }
        }
    };

    DirectAbstraction( NFA &automaton, AbstractionData &abstractionData ) : _automaton( automaton ), _abstractionData( abstractionData )
    {
        _numberOfModules = _abstractionData._abstractModules.size();
    }

    static String projectToModule( const String &state, unsigned moduleIndex )
    {
        static char delimiter[] = " ";
        return state.tokenize( delimiter )[moduleIndex];
    }

    StateSet abstractToConcrete( const String &abstractState )
    {
        if ( _abstractToConcrete.exists( abstractState ) )
        {
            return _abstractToConcrete[abstractState];
        }
        else
        {
            String abstractWithoutProperty = removePropertySuffix( abstractState );
            String propertySuffix = isolatePropertySuffix( abstractState );

            StateSet states;
            states.append( "" );

            for ( unsigned i = 0; i < _numberOfModules; ++i )
            {
                StateSet concreteStatesInModule = _abstractionData._abstractModules[i]._abstractionFunction->abstractToConcrete( projectToModule( abstractWithoutProperty, i ) );
                StateSet newStates;

                for ( unsigned j = 0; j < states.size(); ++j )
                {
                    for ( unsigned k = 0; k < concreteStatesInModule.size(); ++k )
                    {
                        newStates.append( fuseLabels( states[j], concreteStatesInModule[k] ) );
                    }
                }

                states = newStates;
            }

            for ( unsigned i = 0; i < states.size(); ++i )
            {
                states[i] += String( "/" ) + propertySuffix;
            }

            return states;
        }
    }

    // StateSet abstractToConcreteNoSuffix( const String &abstractState )
    // {
    //     if ( _abstractToConcrete.exists( abstractState ) )
    //     {
    //         return _abstractToConcrete[abstractState];
    //     }
    //     else
    //     {
    //         StateSet states;
    //         states.append( "" );

    //         for ( unsigned i = 0; i < _numberOfModules; ++i )
    //         {
    //             StateSet concreteStatesInModule = _abstractionData._abstractModules[i]._abstractionFunction->abstractToConcrete( projectToModule( abstractState, i ) );
    //             StateSet newStates;

    //             for ( unsigned j = 0; j < states.size(); ++j )
    //             {
    //                 for ( unsigned k = 0; k < concreteStatesInModule.size(); ++k )
    //                 {
    //                     newStates.append( fuseLabels( states[j], concreteStatesInModule[k] ) );
    //                 }
    //             }

    //             states = newStates;
    //         }

    //         return states;
    //     }
    // }

    String concreteToAbstract( const String &concreteState )
    {
        if ( _concreteToAbstract.exists( concreteState ) )
        {
            return _concreteToAbstract[concreteState];
        }
        else
        {
            String concreteWithoutProperty = removePropertySuffix( concreteState );
            String propertySuffix = isolatePropertySuffix( concreteState );

            String state = _abstractionData._abstractModules[0]._abstractionFunction->concreteToAbstract( projectToModule( concreteWithoutProperty, 0 ) );

            for ( unsigned i = 1; i < _numberOfModules; ++i )
            {
                state += String( " " ) + _abstractionData._abstractModules[i]._abstractionFunction->concreteToAbstract( projectToModule( concreteWithoutProperty, i ) );
            }

            return state + "/" + propertySuffix;
        }
    }

    // void initalizeDirectAbstraction( const String &state )
    // {
    //     if ( !_abstractToConcrete.exists( state ) )
    //     {
    //         populateMaps( state );
    //     }
    // }

    ThreadState *stateOfModule( unsigned module, const String &composedLabel )
    {
        //        printf( "State of module: %u, %s --> %s\n", module, composedLabel.ascii(), projectToModule( composedLabel, module ).ascii() );
        return &( _abstractionData._abstractModules[module]._concrete->_graph[projectToModule( composedLabel, module )] );
    }

    ThreadState constructStateFromLabel( const String &composedLabel )
    {
        //        printf( "Construct state from label: %s\n", labelWithoutProperty.ascii() );

        String labelWithoutProperty = removePropertySuffix( composedLabel );
        ThreadState result;

        result._label = labelWithoutProperty;

        for ( unsigned i = 0; i < _abstractionData._abstractModules.size(); ++i )
        {
            //            printf( "Module %u: %s\n", i, stateOfModule( i, labelWithoutProperty )._label.ascii() );
            result._requestedEvents += stateOfModule( i, labelWithoutProperty )->_requestedEvents;
            result._blockedEvents += stateOfModule( i, labelWithoutProperty )->_blockedEvents;
        }

        result._requestedEvents.uniqueElements();
        result._blockedEvents.uniqueElements();

        Vector<Event>::iterator it = result._requestedEvents.begin();
        while ( it != result._requestedEvents.end() )
        {
            if ( result._blockedEvents.exists( *it ) )
            {
                result._requestedEvents.erase( it );
            }
            else
            {
                ++it;
            }
        }

        String successor;

        for ( unsigned i = 0; i < result._requestedEvents.size(); ++i )
        {
            Event event = result._requestedEvents[i];

            TS_ASSERT_EQUALS( stateOfModule( 0, labelWithoutProperty )->labelsOfSuccessors( event ).size(), 1U );
            successor = stateOfModule( 0, labelWithoutProperty )->labelsOfSuccessors( event ).first();

            for ( unsigned j = 1; j < _abstractionData._abstractModules.size(); ++ j )
            {
                TS_ASSERT_EQUALS( stateOfModule( j, labelWithoutProperty )->labelsOfSuccessors( event ).size(), 1U );
                successor += String( " " ) + stateOfModule( j, labelWithoutProperty )->labelsOfSuccessors( event ).first();
            }

            result._successors.append( Transition( event, successor ) );
        }

        // Need to complete successors and predecessors

        return result;
    }

    void refineState( const String &stateToRefine, Vector<AbstractStateData> &refinement )
    {
        // printf( "Refine state called: %s\n", stateToRefine.ascii() );
        removeStateFromMapping( stateToRefine );
        addStatesToMapping( refinement );
        adjustAutomaton( stateToRefine, refinement );

    }

    String findConcreteSuccessor( const String &concreteComposedState, const Event &event, String successorSuffix = "" )
    {
        // printf( "Find Concrete Successor called\n" );

        if ( successorSuffix == "" )
        {
            String abstractState = concreteToAbstract( concreteComposedState );
            // printf( "Abstract state: %s\n", abstractState.ascii() );

            String abstractSuccessor = _automaton.getSuccessor( abstractState, eventToLetter( event ) );
            // printf( "Abstract successor: %s\n", abstractSuccessor.ascii() );

            successorSuffix = isolatePropertySuffix( abstractSuccessor );
            // printf( "Successor Suffix: %s\n", successorSuffix.ascii() );
        }

        String stateWithoutProperty = removePropertySuffix( concreteComposedState );

        Map<unsigned, ThreadState *> moduleStates;

        for ( unsigned i = 0; i < _numberOfModules; ++i )
        {
            moduleStates[i] = stateOfModule( i, stateWithoutProperty );
        }

        Map<unsigned, String> moduleSuccessorState;

        for ( unsigned j = 0; j < _numberOfModules; ++j )
        {
            TS_ASSERT_EQUALS( moduleStates[j]->labelsOfSuccessors( event ).size(), 1U );
            moduleSuccessorState[j] = moduleStates[j]->labelsOfSuccessors( event )[0];
        }

        String successorState = moduleSuccessorState[0];
        for ( unsigned j = 1; j < moduleSuccessorState.size(); ++j )
        {
            successorState += String( " " ) + moduleSuccessorState[j];
        }

        return successorState + "/" + successorSuffix;
    }

private:
    Map<String, StateSet> _abstractToConcrete;
    Map<String, String> _concreteToAbstract;

    NFA &_automaton;
    AbstractionData &_abstractionData;
    unsigned _numberOfModules;

    // void populateMaps( const String &abstractState )
    // {
    //     StateSet states;
    //     states.append( "" );

    //     for ( unsigned i = 0; i < _numberOfModules; ++i )
    //     {
    //         StateSet concreteStatesInModule = _abstractionData._abstractModules[i]._abstractionFunction->abstractToConcrete( projectToModule( abstractState, i ) );
    //         StateSet newStates;

    //         for ( unsigned j = 0; j < states.size(); ++j )
    //         {
    //             for ( unsigned k = 0; k < concreteStatesInModule.size(); ++k )
    //             {
    //                 newStates.append( fuseLabels( states[j], concreteStatesInModule[k] ) );
    //             }
    //         }

    //         states = newStates;
    //     }

    //     _abstractToConcrete[abstractState] = states;

    //     for ( unsigned i = 0; i < states.size(); ++i )
    //     {
    //         _concreteToAbstract[states[i]] = abstractState;
    //     }
    // }

    static String fuseLabels( const String &labelOne, const String &labelTwo )
    {
        if ( labelOne == "" )
        {
            return labelTwo;
        }
        else
        {
            return Stringf( "%s %s", labelOne.ascii(), labelTwo.ascii() );
        }
    }

    void removeStateFromMapping( const String &state )
    {
        if ( _abstractToConcrete.exists( state ) )
        {
            _abstractToConcrete.erase( state );
        }
    }

    void addStatesToMapping( Vector<AbstractStateData> &newStates )
    {
        Vector<AbstractStateData>::iterator it;

        for ( it = newStates.begin(); it != newStates.end(); ++it )
        {
            StateSet newStateSet;
            _abstractToConcrete[it->_abstractState] = newStateSet;

            for ( unsigned i = 0; i < it->_concreteStates.size(); ++i )
            {
                _concreteToAbstract[it->_concreteStates.get( i )] = it->_abstractState;
                _abstractToConcrete[it->_abstractState].append( it->_concreteStates.get( i ) );
            }
        }
    }

    void adjustAutomaton( const String &stateToRefine, Vector<AbstractStateData> &refinement )
    {
        adjustAutomatonStates( stateToRefine, refinement );
        adjustEdges( stateToRefine, refinement );
        adjustInitialState();
    }

    void adjustAutomatonStates( const String &stateToRefine, Vector<AbstractStateData> &refinement )
    {
        _automaton.removeState( stateToRefine );

        Vector<AbstractStateData>::iterator it;

        for ( it = refinement.begin(); it != refinement.end(); ++it )
        {
            _automaton.addState( it->_abstractState );
        }
    }

    void adjustEdges( const String &stateToRefine, Vector<AbstractStateData> &refinement )
    {
        adjustPredecessorEdges( stateToRefine );
        adjustSuccessorEdges( refinement, stateToRefine );
    }

    void adjustInitialState()
    {
        String initialState;

        initialState = _abstractionData._abstractModules[0]._concrete->_initialState;

        for ( unsigned i = 1; i < _numberOfModules; ++i )
        {
            initialState += String( " " ) + _abstractionData._abstractModules[i]._concrete->_initialState;
        }

        initialState += String( "/" ) + "q0";

        //printf( "Concrete initial state: %s\n", initialState.ascii() );

        concreteToAbstract( initialState );

        //printf( "Setting initial state to: %s\n", concreteToAbstract( initialState ).ascii() );

        _automaton.setInitialState( concreteToAbstract( initialState ) );
    }

    void adjustPredecessorEdges( const String &stateToRefine )
    {
        Vector<NFA::Edge> predecessorEdges;
        Vector<NFA::Edge>::iterator it = _automaton.edges().begin();

        while ( it != _automaton.edges().end() )
        {
            if ( it->_target == stateToRefine )
            {
                predecessorEdges.append( *it );
                _automaton.edges().erase( it );
            }
            else
            {
                ++it;
            }
        }

        for ( unsigned i = 0; i < predecessorEdges.size(); ++i )
        {
            // printf( "Working on predecessor edge: %s --> (%s) %s\n", predecessorEdges[i]._source.ascii(), predecessorEdges[i]._letter.ascii(), predecessorEdges[i]._target.ascii() );

            String abstractPredecessor = predecessorEdges[i]._source;
            StateSet concretePredecessors = abstractToConcrete( abstractPredecessor );

            for ( unsigned j = 0; j < concretePredecessors.size(); ++j )
            {
                // printf( "Handling concrete predecessor: %s\n", concretePredecessors[j].ascii() );

                if ( eventNotBlockedInConcreteState( concretePredecessors[j], letterToEvent( predecessorEdges[i]._letter ) ) )
                {
                    String concreteSuccessor = findConcreteSuccessor( concretePredecessors[j], letterToEvent( predecessorEdges[i]._letter ), isolatePropertySuffix( stateToRefine ) );

                    // printf( "Concrete successor: %s\n", concreteSuccessor.ascii() );

                    String abstractSuccessor = concreteToAbstract( concreteSuccessor );

                    // printf( "Abstract successor: %s\n", abstractSuccessor.ascii() );

                    if ( !_automaton.edges().exists( NFA::Edge( abstractPredecessor, predecessorEdges[i]._letter, abstractSuccessor ) ) )
                    {
                        _automaton.addEdge( NFA::Edge( abstractPredecessor, predecessorEdges[i]._letter, abstractSuccessor ) );
                    }
                }
            }
        }
    }

    bool eventNotBlockedInConcreteState( const String &concreteState, const Event &event )
    {
        ThreadState state = constructStateFromLabel( concreteState );
        return !state._blockedEvents.exists( event );
    }

    void adjustSuccessorEdges( Vector<AbstractStateData> &refinement, const String &stateToRefine )
    {
        Vector<AbstractStateData>::iterator it;

        for ( it = refinement.begin(); it != refinement.end(); ++it )
        {
            for ( unsigned i = 0; i < it->_concreteStates.size(); ++i )
            {
                addSuccessorEdgesForConcreteState( it->_concreteStates.get( i ), it->_abstractState, stateToRefine );
            }
        }

        Vector<NFA::Edge>::iterator iterator = _automaton.edges().begin();

        while ( iterator != _automaton.edges().end() )
        {
            if ( iterator->_source == stateToRefine )
            {
                _automaton.edges().erase( iterator );
            }
            else
            {
                ++iterator;
            }
        }
    }

    void addSuccessorEdgesForConcreteState( const String &concreteStateLabel, const String &abstractState, const String &stateBeingRefined )
    {
        //        printf( "addSuccessorEdgesForConcreteState called: %s\n", concreteStateLabel.ascii() );

        ThreadState concreteState = constructStateFromLabel( concreteStateLabel );
        //        Vector<Event> allEvents = _abstractionData._composedGraph->_events;

        Vector<Event> requestedEvents = concreteState._requestedEvents;
        Vector<Event> blockedEvents = concreteState._blockedEvents;

        Vector<Event> enabledEvents;

        // printf( "Non blocked events are: " );
        for ( unsigned i = 0; i < requestedEvents.size(); ++i )
        {
            if ( !blockedEvents.exists( requestedEvents[i] ) )
            {
                // printf( "%u ", allEvents[i].type() );
                enabledEvents.append( requestedEvents[i] );
            }
        }
        // printf( "\n" );

        Map<unsigned, ThreadState *> moduleStates;

        for ( unsigned i = 0; i < _numberOfModules; ++i )
        {
            moduleStates[i] = stateOfModule( i, removePropertySuffix( concreteStateLabel ) );
        }

        for ( unsigned i = 0; i < enabledEvents.size(); ++i )
        {
            Map<unsigned, String> moduleSuccessorState;

            for ( unsigned j = 0; j < _numberOfModules; ++j )
            {
                TS_ASSERT_EQUALS( moduleStates[j]->labelsOfSuccessors( enabledEvents[i] ).size(), 1U );
                moduleSuccessorState[j] = moduleStates[j]->labelsOfSuccessors( enabledEvents[i] )[0];
            }

            String successorState = moduleSuccessorState[0];
            for ( unsigned j = 1; j < moduleSuccessorState.size(); ++j )
            {
                successorState += String( " " ) + moduleSuccessorState[j];
            }
            // printf( "Concrete successor state is: %s\n", successorState.ascii() );


            Vector<NFA::Edge>::iterator iterator = _automaton.edges().begin();

            //            printf( "Looking for an edge for source: %s, and event %u\n", abstractState.ascii(), enabledEvents[i].type() );
            bool found = false;
            String suffix;
            while ( ( iterator != _automaton.edges().end() ) && ( !found ) )
            {
                if ( ( iterator->_source == stateBeingRefined ) && ( iterator->_letter == eventToLetter( enabledEvents[i] ) ) )
                {
                    found = true;
                    suffix = isolatePropertySuffix( iterator->_target );
                }
                else
                {
                    ++iterator;
                }
            }

            // printf( "Successor suffix is: %s\n", suffix.ascii() );

            successorState += String( "/" ) + suffix;

            ////////////// LOOK AT THIS FUNCTION
            //// CONCRETE TO ABSTRACT NEEDS TO BE GIVEN THE /q0 as well
            /// ALSO, MAKE SURE ITS OK THE STATE ISNT IN THE MAPPING, BUT NEEDS TO BE PROCESSED
            //            printf( "Source state: %s, event: %s, target state: %s\n", abstractState.ascii(), eventToLetter( enabledEvents[i] ).ascii(), concreteToAbstract( successorState ).ascii() );
            _automaton.addEdge( NFA::Edge( abstractState, eventToLetter( enabledEvents[i] ), concreteToAbstract( successorState ) ) );
        }
    }

    static Event letterToEvent( const String &letter )
    {
        return Event( atoi( letter.ascii() ) );
    }

    static String eventToLetter( const Event &event )
    {
        return Stringf( "%u", event.type() );
    }

    static String removePropertySuffix( const String &state )
    {
        char delimiter[] = "/";
        return state.tokenize( delimiter )[0];
    }

    static String isolatePropertySuffix( const String &state )
    {
        char delimiter[] = "/";
        return state.tokenize( delimiter )[1];
    }
};

#endif // __DirectAbstraction_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
