#ifndef __MultipleThreadIntersector_h__
#define __MultipleThreadIntersector_h__

#include "ManualIntersector.h"
#include "ThreadStateGraph.h"

class MultipleThreadIntersector
{
public:
    MultipleThreadIntersector()
    {
        _temporary = NULL;
        _oldTemporary = NULL;
    }

    ~MultipleThreadIntersector()
    {
        if ( _temporary != NULL )
        {
            delete _temporary;
            _temporary = NULL;
        }
        if ( _oldTemporary != NULL )
        {
            delete _oldTemporary;
            _oldTemporary = NULL;
        }
    }

    void addThread( ThreadStateGraph *thread )
    {
        _inputThreads.append( thread );
    }

    void intersect( ThreadStateGraph &output, char separationChar )
    {
        _separationChar = separationChar;

        if ( _inputThreads.size() == 0 )
        {
            return;
        }
        else if ( _inputThreads.size() == 1 )
        {
            output = *_inputThreads[0];
            return;
        }

        intersectFirstTwoThreads();
        for ( unsigned i = 2; i < _inputThreads.size(); ++i )
        {
            addAnotherThread( _inputThreads[i] );
        }

        output = *_temporary;
    }

private:
    Vector<ThreadStateGraph *> _inputThreads;
    ThreadStateGraph *_temporary;
    ThreadStateGraph *_oldTemporary;
    char _separationChar;

    void intersectFirstTwoThreads()
    {
        _temporary = new ThreadStateGraph;
        singleIntersection( _inputThreads[0], _inputThreads[1], _temporary );
    }

    void singleIntersection( ThreadStateGraph *inputOne, ThreadStateGraph *inputTwo, ThreadStateGraph *output )
    {
        ManualIntersector().intersect( *inputOne, *inputTwo, *output, _separationChar );
        //        printf( "Intersector: number of states after intersection is %u\n", output->_graph.size() );
    }

    void addAnotherThread( ThreadStateGraph *input )
    {
        _oldTemporary = _temporary;
        _temporary = new ThreadStateGraph;
        singleIntersection( _oldTemporary, input, _temporary );

        delete _oldTemporary;
        _oldTemporary = NULL;
    }
};

#endif // __MultipleThreadIntersector_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
