#ifndef __SafetyPatch_h__
#define __SafetyPatch_h__

#include "UniqueVector.h"
#include "Event.h"
#include "NFA.h"

class SafetyPatch
{
public:
    typedef UniqueVector<Event> Events;

    NFA _nfa;
    Map<String, Events> _stateToBlockedEvents;

    void setNfa( NFA &nfa )
    {
        _nfa = nfa;
    }

    void addBlockedEvent( String &state, Event event )
    {
        if ( !_stateToBlockedEvents.exists( state ) )
        {
            Events newEvents;
            _stateToBlockedEvents[state] = newEvents;
        }

        _stateToBlockedEvents[state].append( event );
    }

    void dump()
    {
        printf( "Dumping patch:\n" );

        Map<String, Events>::iterator it;

        for ( it = _stateToBlockedEvents.begin(); it != _stateToBlockedEvents.end(); ++it )
        {
            printf( "State: %s\n", it->first.ascii() );
            for ( unsigned i = 0; i < it->second.size(); ++i )
            {
                printf( "\tBlock %u\n", it->second[i].type() );
            }
            printf( "\n" );
        }
    }
};

#endif // __SafetyPatch_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
