#ifndef __PhilOne_h__
#define __PhilOne_h__

#include "BThread.h"
#include "IEvent.h"
#include "String.h"
#include "T/Truth.h"
#include <cstdio>

// Lefty

class PhilOne : public BThread
{
public:
    PhilOne()
    {
        _eating.add( "Phil#1Eating" );
    }

    void entryPoint()
    {
        while ( T::alwaysTrue() )
        {
            pickLeftFork();

            printf( "Phil #1 picked fork #1 (first)\n" );
            fflush( 0 );

            pickRightFork();

            printf( "Phil #1 picked fork #2 (second)\n" );
            fflush( 0 );

            putLeftDown();

            printf( "Phil #1 put fork #1 (first)\n" );
            fflush( 0 );

            putRightDown();

            printf( "Phil #1 put fork #2 (second)\n" );
            fflush( 0 );
        }
    }

private:
    AtomicPropositions _eating;

    void clearAll()
    {
        _requested.clear();
        _watched.clear();
        _blocked.clear();
    }

    void pickLeftFork()
    {
        clearAll();
        _requested.append( IEvent::PHIL_ONE_PICK_LEFT );
        _blocked = blockAllBut( IEvent::PHIL_ONE_PICK_LEFT );
        bSync( _requested, _watched, _blocked, Stringf( "(1,1)" ) );
    }

    void pickRightFork()
    {
        clearAll();
        _requested.append( IEvent::PHIL_ONE_PICK_RIGHT );
        _blocked = blockAllBut( IEvent::PHIL_ONE_PICK_RIGHT );
        bSync( _requested, _watched, _blocked, Stringf( "(1,2)" ) );
    }

    void putLeftDown()
    {
        clearAll();
        _requested.append( IEvent::PHIL_ONE_PUT_LEFT );
        _blocked = blockAllBut( IEvent::PHIL_ONE_PUT_LEFT );
        bSync( _requested, _watched, _blocked, _eating, Stringf( "(1,3)" ) );
    }

    void putRightDown()
    {
        clearAll();
        _requested.append( IEvent::PHIL_ONE_PUT_RIGHT );
        _blocked = blockAllBut( IEvent::PHIL_ONE_PUT_RIGHT );
        bSync( _requested, _watched, _blocked, Stringf( "(1,4)" ) );
    }

    Vector<Event> blockAllBut( IEvent::EventType type )
    {
        Vector<Event> blocked;

        if ( type != IEvent::PHIL_ONE_PICK_LEFT )
        {
            blocked.append( IEvent::PHIL_ONE_PICK_LEFT );
        }
        if ( type != IEvent::PHIL_ONE_PICK_RIGHT )
        {
            blocked.append( IEvent::PHIL_ONE_PICK_RIGHT );
        }
        if ( type != IEvent::PHIL_ONE_PUT_LEFT )
        {
            blocked.append( IEvent::PHIL_ONE_PUT_LEFT );
        }
        if ( type != IEvent::PHIL_ONE_PUT_RIGHT )
        {
            blocked.append( IEvent::PHIL_ONE_PUT_RIGHT );
        }

        return blocked;
    }
};

#endif // __PhilOne_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
