#ifndef __PhilThree_h__
#define __PhilThree_h__

#include "BThread.h"
#include "IEvent.h"
#include "String.h"
#include "T/Truth.h"
#include <cstdio>

// Righty

class PhilThree : public BThread
{
public:
    void entryPoint()
    {
        while ( T::alwaysTrue() )
        {
            pickRightFork();

            printf( "Phil #3 picked fork #1 (first)\n" );
            fflush( 0 );

            pickLeftFork();

            printf( "Phil #3 picked fork #3 (second)\n" );
            fflush( 0 );

            putRightDown();

            printf( "Phil #3 put fork #1 (first)\n" );
            fflush( 0 );

            putLeftDown();

            printf( "Phil #3 put fork #3 (second)\n" );
            fflush( 0 );
        }
    }

private:
    AtomicPropositions _eating;

    void clearAll()
    {
        _requested.clear();
        _watched.clear();
        _blocked.clear();
    }

    void pickRightFork()
    {
        clearAll();
        _requested.append( IEvent::PHIL_THREE_PICK_RIGHT );
        _blocked = blockAllBut( IEvent::PHIL_THREE_PICK_RIGHT );
        bSync( _requested, _watched, _blocked, Stringf( "(3,1)" ) );
    }

    void pickLeftFork()
    {
        clearAll();
        _requested.append( IEvent::PHIL_THREE_PICK_LEFT );
        _blocked = blockAllBut( IEvent::PHIL_THREE_PICK_LEFT );
        bSync( _requested, _watched, _blocked, Stringf( "(3,2)" ) );
    }

    void putRightDown()
    {
        clearAll();
        _requested.append( IEvent::PHIL_THREE_PUT_RIGHT );
        _blocked = blockAllBut( IEvent::PHIL_THREE_PUT_RIGHT );
        bSync( _requested, _watched, _blocked, Stringf( "(3,3)" ) );
    }

    void putLeftDown()
    {
        clearAll();
        _requested.append( IEvent::PHIL_THREE_PUT_LEFT );
        _blocked = blockAllBut( IEvent::PHIL_THREE_PUT_LEFT );
        bSync( _requested, _watched, _blocked, Stringf( "(3,4)" ) );
    }

    Vector<Event> blockAllBut( IEvent::EventType type )
    {
        Vector<Event> blocked;

        if ( type != IEvent::PHIL_THREE_PICK_LEFT )
        {
            blocked.append( IEvent::PHIL_THREE_PICK_LEFT );
        }
        if ( type != IEvent::PHIL_THREE_PICK_RIGHT )
        {
            blocked.append( IEvent::PHIL_THREE_PICK_RIGHT );
        }
        if ( type != IEvent::PHIL_THREE_PUT_LEFT )
        {
            blocked.append( IEvent::PHIL_THREE_PUT_LEFT );
        }
        if ( type != IEvent::PHIL_THREE_PUT_RIGHT )
        {
            blocked.append( IEvent::PHIL_THREE_PUT_RIGHT );
        }

        return blocked;
    }
};

#endif // __PhilThree_h__

//
// Local Variables:
// compile-command: "make -C .. "
// tags-file-name: "../TAGS"
// c-basic-offset: 4
// End:
//
